import logging
from typing import Optional, Awaitable, Any, Union

import tornado.web
from tornado import httputil

from middlewares.base import MiddlewareType

logger = logging.getLogger(__name__)


class BaseHandler(tornado.web.RequestHandler):
    """A class to collect common handler methods - all other handlers should
    subclass this one.
    """

    def __init__(self, application: "Application", request: httputil.HTTPServerRequest, **kwargs: Any) -> None:
        super().__init__(application, request, **kwargs)

        # 拷贝一份中间件列表
        self.middleware_list = self.application.middlewares

    def prepare(self) -> Optional[Awaitable[None]]:
        """
        在每个请求的最开始的时候被调用, 在 get/post/等方法之前.
        """
        super().prepare()
        return self._process_middlewares(mw_type=MiddlewareType.BEFORE_REQUEST)

    def on_finish(self) -> None:
        """在一个请求结束后被调用."""
        super().on_finish()
        self._process_middlewares(mw_type=MiddlewareType.FINISHED)

    def finish(self, chunk: Union[str, bytes, dict] = None) -> "Future[None]":
        """完成响应后调用."""
        self._process_middlewares(MiddlewareType.AFTER_RESPONSE, chunk)
        return super().finish(chunk)

    def _process_middlewares(self, mw_type, *args, **kwargs):
        for middleware in self.middleware_list:
            try:
                if middleware.mw_type is mw_type:
                    should_run_method = getattr(middleware, 'should_run', None)
                    if should_run_method and callable(should_run_method):
                        should_run = should_run_method(self, *args, **kwargs)  # 执行中间件的should_run方法
                        if should_run:
                            run_method = getattr(middleware, 'run', None)
                            if run_method and callable(run_method):
                                run_method(self, *args, **kwargs)  # 执行中间件的run方法
            except Exception as e:
                logger.exception(e)

    def data_received(self, chunk: bytes) -> Optional[Awaitable[None]]:
        pass



import os
import signal
import time

import tornado.escape
import tornado.ioloop
import tornado.locks
import tornado.web

from .log import loquat_logger

# Make filepaths relative to settings.
here = os.path.dirname(os.path.abspath(__file__))
pardir = os.path.abspath(os.path.join(here, os.path.pardir))


class DeploymentType:
    SOLO = "SOLO"  # 也就是本地模式
    DEV = "DEV"  # 开发环境
    QA = "QA"  # 测试环境
    UAT = "UAT"  # 预生产环境
    PROD = "PROD"  # 生产环境
    dict = {
        SOLO: "SOLO",
        DEV: "DEV",
        QA: "QA",
        UAT: "UAT",
        PROD: "PROD"
    }


class AppConfig(object):
    def __init__(self, *args, **kwargs):
        self.app_name = ''
        self.port = 8000
        self.env = 'SOLO'
        self.handlers = []
        self.app_settings = {}
        self.app_properties = {}
        self.middlewares = []

        if kwargs:
            self.set_config(**kwargs)

    def set_config(self, **kwargs):

        config = kwargs.copy()

        try:
            self.app_name = config['app_name']
            del config['app_name']
        except KeyError:
            pass  # Leave what was set or default

        try:
            self.port = config['port']
            del config['port']
        except KeyError:
            pass  # Leave what was set or default

        try:
            self.env = config['env']
            del config['env']
        except KeyError:
            pass

        try:
            self.handlers = config['handlers']
            del config['handlers']
        except KeyError:
            pass

        try:
            self.app_settings = config['app_settings']
            del config['app_settings']
        except KeyError:
            pass

        try:
            app_properties = config['app_properties'] if 'app_properties' in config.keys() else {}
            self.app_properties = {**self.app_properties, **app_properties}
            del config['app_properties']
        except KeyError:
            pass

        try:
            self.middlewares = config['middlewares']
            del config['middlewares']
        except KeyError:
            pass


class Application(tornado.web.Application):
    """Loquat Application"""

    def __init__(self, app_config: AppConfig):

        super(Application, self).__init__(app_config.handlers, **app_config.app_settings)

        self.app_config = app_config  # 赋值app_config

        self._init_middlewares()  # 初始化中间件

        self._init_app_properties()  # 设置application的属性

        loquat_logger.debug('Inited Loquat Application')

    def _init_app_properties(self):
        """
        设置application的属性
        """
        for name in self.app_config.app_properties.keys():
            self.__setattr__(name, self.app_config.app_properties.get(name))

    def _init_middlewares(self):
        """
        初始化中间件
        """
        self.middlewares = []
        for middleware_cls in self.app_config.middlewares:
            middleware_instance = middleware_cls()
            self.middlewares.append(middleware_instance)
        self.middlewares = sorted(self.middlewares, key=lambda m: m.mw_order)

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)


def shutdown_sig_handler(sig, frame):
    """handle shutdown signal"""

    loquat_logger.warning('caught signal: %s', sig)

    max_wait_seconds_before_shutdown = 1

    def shutdown():
        loquat_logger.info('stopping http server')
        server.stop()

        loquat_logger.info('will shutdown in %s seconds...', max_wait_seconds_before_shutdown)
        io_loop = tornado.ioloop.IOLoop.instance()

        deadline = time.time() + max_wait_seconds_before_shutdown

        def stop_loop():
            now = time.time()
            if now < deadline:
                io_loop.add_timeout(now + 1, stop_loop)
            else:
                loquat_logger.info('shutdown')
                io_loop.stop()

        stop_loop()

    tornado.ioloop.IOLoop.instance().add_callback(shutdown)


def run(app_config: AppConfig):
    """run web application"""

    global server
    if not app_config:
        app_config = AppConfig()

    application = Application(app_config)

    server = tornado.httpserver.HTTPServer(application, xheaders=True)
    server.listen(app_config.port)

    loquat_logger.info("%s started on port %s." % (app_config.app_name, app_config.port,))

    signal.signal(signal.SIGTERM, shutdown_sig_handler)
    signal.signal(signal.SIGINT, shutdown_sig_handler)

    tornado.ioloop.IOLoop.instance().start()



from abc import ABCMeta, abstractmethod


class MiddlewareType:
    BEFORE_REQUEST = 'BEFORE_REQUEST'  # 在request之前
    AFTER_RESPONSE = 'AFTER_RESPONSE'  # 在response之后
    FINISHED = 'FINISHED'  # 在请求结束之后
    dict = {
        BEFORE_REQUEST: "BEFORE_REQUEST",
        AFTER_RESPONSE: "AFTER_RESPONSE",
        FINISHED: "FINISHED"
    }


class BaseMiddleware(object, metaclass=ABCMeta):

    def __init__(self, mw_order=0, mw_type=MiddlewareType.BEFORE_REQUEST):

        self._mw_order = mw_order
        self._mw_type = mw_type

    @property
    def mw_order(self):
        """middleware的执行顺序，值越小越靠前执行
        """
        return self._mw_order

    @mw_order.setter
    def mw_order(self, value):
        if not isinstance(value, int):
            raise ValueError('value must be an integer!')
        if value < 0:
            raise ValueError('value must bigger than 0!')
        self._mw_order = value

    @property
    def mw_type(self):
        """middleware的类型。有：MiddlewareType.FINISHED, MiddlewareType.AFTER_RESPONSE, MiddlewareType.BEFORE_REQUEST
        """
        return self._mw_type

    @mw_type.setter
    def mw_type(self, value):

        if value not in [MiddlewareType.FINISHED, MiddlewareType.AFTER_RESPONSE, MiddlewareType.BEFORE_REQUEST]:
            raise ValueError(
                'value must in [MiddlewareType.FINISHED, MiddlewareType.AFTER_RESPONSE, MiddlewareType.BEFORE_REQUEST]')

        self._mw_type = value

    @abstractmethod
    def should_run(self, handler, *args, **kwargs) -> bool:
        """
        是否需要运行中间件。True为运行，False为不运行
        @param handler: tornado.web.RequestHandler的子类
        @param args:
        @param kwargs:
        """
        pass

    @abstractmethod
    def run(self, handler, *args, **kwargs) -> None:
        """
        运行中间件
        @param handler: tornado.web.RequestHandler的子类
        @param args:
        @param kwargs:
        """
        pass



import copy
from typing import Any, Dict

unicode_type = str
basestring_type = str

copy_list = (lambda lb: copy(lb) if lb else [])


class ObjectDict(Dict[str, Any]):
    """Makes a dictionary behave like an object, with attribute-style access.
    """

    def __getattr__(self, name: str) -> Any:
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name: str, value: Any) -> None:
        self[name] = value

    def __delattr__(self, name):
        try:
            del self[name]
        except KeyError:
            raise AttributeError(name)


def list_to_str(value, encode=None):
    """recursively convert list content into string

    :arg list value: The list that need to be converted.
    :arg function encode: Function used to encode object.
    """
    result = []
    for index, v in enumerate(value):
        if isinstance(v, dict):
            result.append(dict_to_str(v, encode))
            continue

        if isinstance(v, list):
            result.append(list_to_str(v, encode))
            continue

        if encode:
            result.append(encode(v))
        else:
            result.append(v)

    return result


def dict_to_str(origin_value, encode=None):
    """recursively convert dict content into string
    """
    value = copy.deepcopy(origin_value)
    for k, v in value.items():
        if isinstance(v, dict):
            value[k] = dict_to_str(v, encode)
            continue

        if isinstance(v, list):
            value[k] = list_to_str(v, encode)
            continue

        if encode:
            value[k] = encode(v)
        else:
            value[k] = v

    return value


def camel_to_underscore(name):
    """
    convert CamelCase style to under_score_case
    """
    as_list = []
    length = len(name)
    for index, i in enumerate(name):
        if index != 0 and index != length - 1 and i.isupper():
            as_list.append('_%s' % i.lower())
        else:
            as_list.append(i.lower())

    return ''.join(as_list)


def import_object(name: str) -> Any:
    """Imports an object by name.

    ``import_object('x')`` is equivalent to ``import x``.
    ``import_object('x.y.z')`` is equivalent to ``from x.y import z``.

    >>> import tornado.escape
    >>> import_object('tornado.escape') is tornado.escape
    True
    >>> import_object('tornado.escape.utf8') is tornado.escape.utf8
    True
    >>> import_object('tornado') is tornado
    True
    >>> import_object('tornado.missing_module')
    Traceback (most recent call last):
        ...
    ImportError: No module named missing_module
    """
    if name.count(".") == 0:
        return __import__(name)

    parts = name.split(".")
    obj = __import__(".".join(parts[:-1]), fromlist=[parts[-1]])
    try:
        return getattr(obj, parts[-1])
    except AttributeError:
        raise ImportError("No module named %s" % parts[-1])

# class Convert(object):
#     __slots__ = ['list_to_str', 'dict_to_str']
#
#     def __init__(self) -> None:
#         super().__init__()
#
#     def __getattr__(self, name):
#         if name in self.__slots__:
#             return globals().get(name)
#
#         raise AttributeError('Convert has no attribute %s' % name)
#
#
# convert = Convert()
